package org.artifactory.engine;

import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.ConstantValues;
import org.artifactory.fs.RepoResource;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.request.InternalRequestContext;
import org.artifactory.request.Request;
import org.artifactory.resource.FileResource;
import org.artifactory.resource.ResourceStreamHandle;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.traffic.TrafficService;
import org.artifactory.traffic.v2.TrafficEntryV2;
import org.jfrog.storage.binstore.providers.base.RedirectType;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.IOException;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertEquals;

public class RedirectTrafficLoggingTest extends ArtifactoryHomeBoundTest {

    @DataProvider(name = "data")
    public Object[][] data() {
        return new Object[][] {
                {
                        false, "", TrafficEntryV2.TrafficActor.USER, "a:b", 7, 7, 7,
                        TrafficEntryV2.TrafficTarget.S3, TrafficEntryV2.TrafficAction.REDIRECT
                },
                {
                        false, "Range: bytes=0-1", TrafficEntryV2.TrafficActor.USER, "a:b", 2, 7, 2,
                        TrafficEntryV2.TrafficTarget.S3, TrafficEntryV2.TrafficAction.REDIRECT
                },
                {
                        true, "", TrafficEntryV2.TrafficActor.USER, "a:b", 7, 7, 7,
                        TrafficEntryV2.TrafficTarget.S3, TrafficEntryV2.TrafficAction.REDIRECT
                },
                {
                        true, "Range: bytes=0-1", TrafficEntryV2.TrafficActor.USER, "a:b", 7, 7, 7,
                        TrafficEntryV2.TrafficTarget.S3, TrafficEntryV2.TrafficAction.REDIRECT
                }
        };
    }

    @BeforeMethod
    public void setup() {
        ArtifactoryContext artifactoryContext = Mockito.mock(ArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
    }

    @AfterMethod
    public void afterMethod() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test(dataProvider = "data")
    public void testName(
            boolean ignoreRangeHeader,
            String rangeHeader,
            TrafficEntryV2.TrafficActor actor,
            String repoPathStr,
            int requestedSize,
            int size,
            int transferredSize,
            TrafficEntryV2.TrafficTarget target,
            TrafficEntryV2.TrafficAction action
    ) throws IOException {
        homeStub.getArtifactoryProperties()
                .setProperty(ConstantValues.trafficIgnoreRangeHeaderWhenCalcBinarySizeForRedirect.getPropertyName(),
                        Boolean.toString(ignoreRangeHeader));

        AuthorizationService authorizationService = mock(AuthorizationService.class);
        TrafficService trafficService = mock(TrafficService.class);
        DownloadServiceImpl downloadService = new DownloadServiceImpl();
        downloadService.trafficService = trafficService;
        downloadService.authorizationService = authorizationService;
        ArtifactoryResponse response = mock(ArtifactoryResponse.class);
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoPath repoPath = new RepoPathImpl("a", "b");
        FileInfoImpl fileInfo = new FileInfoImpl(repoPath);
        fileInfo.setSize("content".length());
        RepoResource resource = new FileResource(fileInfo);
        ResourceStreamHandle handle = mock(ResourceStreamHandle.class);
        Request request = mock(Request.class);
        if (!rangeHeader.isEmpty()) {
            when(request.getHeader("Range")).thenReturn(rangeHeader);
        }
        when(requestContext.getRequest()).thenReturn(request);
        Repo responseRepo = mock(Repo.class);

        downloadService.sendStreamOrRedirect(requestContext, response, resource, repoPath, handle,
                "signedUrl", request, responseRepo, RedirectType.S3);

        assertCapturedTrafficEntry(trafficService, actor, repoPathStr, requestedSize, size, transferredSize, target,
                action);
    }

    protected void assertCapturedTrafficEntry(TrafficService trafficService, TrafficEntryV2.TrafficActor actor,
            String repoPathStr, int requestedSize, int size, int transferredSize, TrafficEntryV2.TrafficTarget target,
            TrafficEntryV2.TrafficAction action) {
        ArgumentCaptor<TrafficEntryV2> captor = ArgumentCaptor.forClass(TrafficEntryV2.class);
        verify(trafficService, times(1)).handleTrafficEntry(captor.capture());
        TrafficEntryV2 actual = captor.getValue();

        assertTrafficEntry(actual, actor, repoPathStr, requestedSize, size, transferredSize, target, action);
    }

    protected void assertTrafficEntry(TrafficEntryV2 actual, TrafficEntryV2.TrafficActor actor, String repoPathStr,
            int requestedSize, int size, int transferredSize, TrafficEntryV2.TrafficTarget target,
            TrafficEntryV2.TrafficAction action) {
        assertEquals(actual.getAction(), action);
        assertEquals(actual.getActor(), actor);
        assertEquals(actual.getRedirectTarget(), target);
        assertEquals(actual.getRepoPath(), repoPathStr);
        assertEquals(actual.getRequestedSize(), requestedSize);
        assertEquals(actual.getSize(), size);
        assertEquals(actual.getTransferredSize(), transferredSize);
    }

}
