/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.build.dao;

import com.google.common.collect.Lists;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.build.BuildProps;
import org.artifactory.api.build.GeneralBuild;
import org.artifactory.api.build.ModuleArtifact;
import org.artifactory.api.build.ModuleDependency;
import org.artifactory.api.build.PublishedModule;
import org.artifactory.api.build.model.diff.BuildParams;
import org.artifactory.api.jackson.JacksonReader;
import org.artifactory.api.rest.build.ContinueBuildFilter;
import org.artifactory.build.BuildId;
import org.artifactory.build.BuildNameAndRepo;
import org.artifactory.build.BuildRun;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.build.dao.utils.BuildsDaoUtils;
import org.artifactory.storage.db.build.entity.BuildEntity;
import org.artifactory.storage.db.build.entity.BuildEntityRecord;
import org.artifactory.storage.db.build.entity.BuildIdEntity;
import org.artifactory.storage.db.build.entity.BuildPromotionStatus;
import org.artifactory.storage.db.build.entity.BuildProperty;
import org.artifactory.storage.db.build.service.BuildRunImpl;
import org.artifactory.storage.db.build.service.BuildStoreServiceImpl;
import org.artifactory.storage.db.util.BaseDao;
import org.artifactory.storage.db.util.JdbcHelper;
import org.artifactory.storage.db.util.querybuilder.ArtifactoryQueryWriter;
import org.artifactory.util.StreamUtils;
import org.jfrog.storage.util.DbUtils;
import org.jfrog.storage.util.querybuilder.QueryWriter;
import org.jfrog.storage.wrapper.BlobWrapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.validation.constraints.NotNull;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import static org.jfrog.storage.util.DbUtils.tableExists;

/**
 * Date: 10/30/12
 * Time: 12:44 PM
 *
 * @author freds
 */
@Slf4j
@Repository
public class BuildsDao extends BaseDao {

    private static final int PROP_VALUE_MAX_SIZE = 2048;
    private static final String DELETE_JSON_BY_BUILD_ID = "DELETE FROM build_jsons WHERE build_id=?";
    private static final String BUILD_JSONS_TABLE = "build_jsons";
    private static final String BUILD_NAME_FIELD = "build_name";
    private static final String BUILD_DATE_FIELD = "build_date";
    private static final String AND = " AND ";

    private final ArtifactoryDbProperties dbProperties;
    private final DbService dbService;

    @Autowired
    public BuildsDao(JdbcHelper jdbcHelper, ArtifactoryDbProperties dbProperties, DbService dbService) {
        super(jdbcHelper);
        this.dbProperties = dbProperties;
        this.dbService = dbService;
    }

    public int createBuild(BuildEntity build, BlobWrapper jsonBlob) throws SQLException {
        int res = createBuild(build);
        res += createBuildJsonIfNeeded(build.getBuildId(), jsonBlob, false);
        res = createBuildProps(build, res);
        res = createPromotions(build, res);
        return res;
    }

    public int createBuild(BuildEntity b) throws SQLException {
        return jdbcHelper.executeUpdate("INSERT INTO builds " +
                        "(build_id, " +
                        "build_name, build_number, build_date, " +
                        "ci_url, created, created_by, " +
                        "modified, modified_by, repo) " +
                        "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                b.getBuildId(),
                b.getBuildName(), b.getBuildNumber(), b.getBuildDate(),
                b.getCiUrl(), b.getCreated(), b.getCreatedBy(),
                nullIfZero(b.getModified()), b.getModifiedBy(), b.getRepo());
    }

    private int createBuildProps(BuildEntity b, int res) throws SQLException {
        int maxPropValue = getDbIndexedValueMaxSize(PROP_VALUE_MAX_SIZE);
        int nbProps = b.getProperties().size();
        if (nbProps != 0) {
            for (BuildProperty bp : b.getProperties()) {
                String propValue = bp.getPropValue();
                if (propValue != null && propValue.length() > maxPropValue) {
                    log.info("Trimming property value to {} characters {}", maxPropValue, bp.getPropKey());
                    log.debug("Trimming property value to {} characters {}: {}", maxPropValue, bp.getPropKey(),
                            bp.getPropValue());
                    propValue = StringUtils.substring(propValue, 0, maxPropValue);
                }
                res += jdbcHelper.executeUpdate("INSERT INTO build_props " +
                                "(prop_id, build_id, prop_key, prop_value) " +
                                "VALUES (?,?,?,?)",
                        bp.getPropId(), bp.getBuildId(), bp.getPropKey(), propValue);
            }
        }
        return res;
    }

    private int createPromotions(BuildEntity b, int res) throws SQLException {
        int nbPromotions = b.getPromotions().size();
        if (nbPromotions != 0) {
            for (BuildPromotionStatus bp : b.getPromotions()) {
                res += jdbcHelper.executeUpdate("INSERT INTO build_promotions " +
                                "(build_id, created, created_by, status, repo, promotion_comment, ci_user) " +
                                "VALUES (?,?,?,?,?,?,?)",
                        bp.getBuildId(), bp.getCreated(), bp.getCreatedBy(),
                        bp.getStatus(), bp.getRepository(), bp.getComment(), bp.getCiUser());
            }
        }
        return res;
    }

    public int addPromotionStatus(long buildId, BuildPromotionStatus promotionStatus, BlobWrapper jsonBlob,
            String currentUser, long currentTime) throws SQLException {
        int res = jdbcHelper.executeUpdate("UPDATE builds SET" +
                " modified = ?, modified_by = ?" +
                " WHERE build_id = ?", currentTime, currentUser, buildId);
        res += createBuildJsonIfNeeded(buildId, jsonBlob, true);
        res += jdbcHelper.executeUpdate("INSERT INTO build_promotions " +
                        "(build_id, created, created_by, status, repo, promotion_comment, ci_user) " +
                        "VALUES (?,?,?,?,?,?,?)",
                promotionStatus.getBuildId(), promotionStatus.getCreated(), promotionStatus.getCreatedBy(),
                promotionStatus.getStatus(), promotionStatus.getRepository(), promotionStatus.getComment(),
                promotionStatus.getCiUser());
        return res;
    }

    /**
     * In legacy mode (up until the converter finishes) all nodes write both into database and repo.
     * This means adding/replacing in the build jsons table is still required
     */
    private int createBuildJsonIfNeeded(long buildId, BlobWrapper jsonBlob, boolean delete) throws SQLException {
        int res = 0;
        if (jsonBlob != null) {
            if (delete) {
                res += jdbcHelper.executeUpdate(DELETE_JSON_BY_BUILD_ID, buildId);
            }
            res += jdbcHelper.executeUpdate("INSERT INTO build_jsons (build_id, build_info_json) VALUES(?, ?)", buildId,
                    jsonBlob);
        }
        return res;
    }

    public int deleteAllBuilds(boolean legacy) throws SQLException {
        int res = jdbcHelper.executeUpdate("DELETE FROM build_props");
        if (legacy) {
            res += jdbcHelper.executeUpdate("DELETE FROM build_jsons");
        }
        res += jdbcHelper.executeUpdate("DELETE FROM build_promotions");
        res += jdbcHelper.executeUpdate("DELETE FROM builds");
        return res;
    }

    public int deleteBuild(long buildId, boolean legacy) throws SQLException {
        int res = jdbcHelper.executeUpdate("DELETE FROM build_props WHERE build_id=?", buildId);
        if (legacy) {
            res += jdbcHelper.executeUpdate(DELETE_JSON_BY_BUILD_ID, buildId);
        }
        res += jdbcHelper.executeUpdate("DELETE FROM build_promotions WHERE build_id=?", buildId);
        res += jdbcHelper.executeUpdate("DELETE FROM builds WHERE build_id=?", buildId);
        return res;
    }

    public <T> T getJsonBuild(long buildId, Class<T> clazz) throws SQLException {
        ResultSet rs = null;
        InputStream jsonStream = null;
        try {
            rs = jdbcHelper.executeSelect("SELECT build_info_json FROM build_jsons WHERE build_id = ?", buildId);
            if (rs.next()) {
                jsonStream = rs.getBinaryStream(1);
                if (CharSequence.class.isAssignableFrom(clazz)) {
                    //noinspection unchecked
                    return (T) IOUtils.toString(jsonStream, StandardCharsets.UTF_8.name());
                }
                return JacksonReader.streamAsClass(jsonStream, clazz);
            }
        } catch (IOException e) {
            throw new SQLException("Failed to read JSON data for build '" + buildId + "' due to: " + e.getMessage(), e);
        } finally {
            DbUtils.close(rs);
            StreamUtils.close(jsonStream);
        }
        return null;
    }

    public BuildEntity getBuild(long buildId) throws SQLException {
        BuildEntityRecord build = getBuildEntityRecord(buildId);
        if (build == null) {
            return null;
        }
        return new BuildEntity(build,
                findBuildProperties(build.getBuildId()),
                findBuildPromotions(build.getBuildId()));
    }

    /**
     * return only the  build record without properties which aren't set yet.
     */
    private BuildEntityRecord getBuildEntityRecord(long buildId) throws SQLException {
        BuildEntity build = null;
        try (ResultSet rs = jdbcHelper.executeSelect("SELECT * FROM builds WHERE build_id = ?", buildId)) {
            if (rs.next()) {
                build = resultSetToBuild(rs);
            }
        }
        return build;
    }

    /**
     * returns a list of previous builds
     */
    public List<GeneralBuild> getPrevBuildsList(String buildName, String buildDate, String buildRepo) {
        String buildsQuery = "SELECT builds.*, build_promotions.status " +
                "FROM builds " +
                "LEFT JOIN build_promotions on build_promotions.build_id = builds.build_id  " +
                "WHERE builds.build_name = ? and builds.repo = ? and builds.build_date < ? " +
                "ORDER BY build_number DESC";

        List<GeneralBuild> buildList = new ArrayList<>();
        try (ResultSet rs = jdbcHelper.executeSelect(buildsQuery, buildName, buildRepo, Long.parseLong(buildDate))) {
            while (rs.next()) {
                GeneralBuild buildEntity = resultSetToGeneralBuild(rs, rs.getLong(1));
                buildEntity.setStatus(rs.getString(10));
                buildList.add(buildEntity);
            }
        } catch (Exception e) {
            log.error("Failed to get previous builds list, {}", e.getMessage());
            log.debug("Failed to get previous builds list", e);
        }
        return buildList;
    }

    public List<GeneralBuild> getBuildForName(String buildName, String buildRepo,
            ContinueBuildFilter continueBuildFilter) throws SQLException {
        List<GeneralBuild> buildList;
        if (continueBuildFilter.getLimit() <= 0) {
            return Collections.emptyList();
        }
        List<Object> params = new ArrayList<>();
        QueryWriter buildsQuery = new QueryWriter(dbService.getDatabaseType());
        buildsQuery.select("builds.* , \n" +
                "         '0' as module_cnt ,\n" +
                "         '0' as artifact_cnt,\n" +
                "         '0' as dependency_cnt ")
                .from("builds ")
                .orderBy("build_date DESC, build_number DESC")
                .where("build_name = ? AND repo = ?")
                .limit(continueBuildFilter.getLimit());
        params.add(buildName);
        params.add(buildRepo);

        BuildId continueBuild = continueBuildFilter.getContinueBuildId();
        if (continueBuild != null) {
            String whereAddition = " AND (build_date < ? OR (build_date = ? AND build_number < ? )) ";
            buildsQuery.where("build_name = ? AND repo = ? " + whereAddition);

            params.add(continueBuild.getStartedDate().getTime());
            params.add(continueBuild.getStartedDate().getTime());
            params.add(continueBuild.getNumber());
        }

        try (ResultSet rs = jdbcHelper.executeSelect(buildsQuery.build(), params.toArray())) {
            buildList = resultSetToGeneralBuildList(rs);
        }
        return Optional.of(buildList).orElse(Collections.emptyList());
    }

    private List<GeneralBuild> resultSetToGeneralBuildList(ResultSet rs) throws SQLException {
        List<GeneralBuild> answer = new ArrayList<>();
        while (rs.next()) {
            Long id = rs.getLong("build_id");
            GeneralBuild buildEntity = resultSetToGeneralBuild(rs, id);
            buildEntity.setNumOfModules(Integer.toString(rs.getInt("module_cnt")));
            buildEntity.setNumOfArtifacts(Integer.toString(rs.getInt("artifact_cnt")));
            buildEntity.setNumOfDependencies(Integer.toString(rs.getInt("dependency_cnt")));
            answer.add(buildEntity);
        }
        return answer;
    }

    /**
     * get Module Artifact diff with paging
     */
    public List<ModuleArtifact> getModuleArtifactsForDiffWithPaging(BuildParams buildParams,
            boolean onlyDefaultBuildRepoExists) {
        List<ModuleArtifact> artifacts = new ArrayList<>();
        Object[] diffParams = getBuildDependencyParams(buildParams);
        try (ResultSet rs = jdbcHelper.executeSelect(BuildQueries.MODULE_ARTIFACT_DIFF_QUERY, diffParams)) {
            while (rs.next()) {
                ModuleArtifact artifact =
                        new ModuleArtifact(null, null, rs.getString(1), rs.getString(2), rs.getString(3));
                artifact.setStatus(rs.getString(4));
                artifacts.add(artifact);
            }
            // update artifact repo path data
            if (!artifacts.isEmpty()) {
                final Map<String, ModuleArtifact> artifactNodes =
                        getArtifactNodes(buildParams.getBuildName(), buildParams.getCurrBuildNum(),
                                buildParams.getCurrBuildDate(), onlyDefaultBuildRepoExists);
                for (ModuleArtifact artifact : artifacts) {
                    ModuleArtifact moduleArtifact = artifactNodes.get(artifact.getSha1());
                    if (moduleArtifact != null) {
                        artifact.setRepoKey(moduleArtifact.getRepoKey());
                        artifact.setPath(moduleArtifact.getPath());
                    }
                }
            }
        } catch (SQLException e) {
            log.error("Failed to get module artifacts with paging, {}", e.getMessage());
            log.debug("Failed to get module artifacts with paging", e);
        }
        return artifacts;
    }

    /**
     * get BuildProps through distinct filter.
     */
    public List<BuildProps> getBuildPropsList(BuildParams buildParams) {
        String buildPropsQuery;
        if (buildParams.isEnvProps()) {
            buildPropsQuery = BuildQueries.BUILD_ENV_PROPS_BY_BUILD_ID;
        } else {
            buildPropsQuery = BuildQueries.BUILD_SYSTEM_PROPS_BY_BUILD_ID;
        }
        Long buildId = getBuildId(buildParams);
        return getBuildPropsList(buildPropsQuery, buildId);
    }

    /**
     * get build props - distinct programmatically.
     * would be slow on large list but several thousands should be fine.
     */
    public List<BuildProps> getBuildPropsList(String buildPropsQuery, Long buildId) {
        List<BuildProps> buildPropsList = new ArrayList<>();
        if (buildId == null) { return buildPropsList; }

        try (ResultSet rs = jdbcHelper.executeSelect(buildPropsQuery, buildId)) {
            while (rs.next()) {
                buildPropsList.add(resultSetToBuildProps(rs));
            }
        } catch (SQLException e) {
            log.error("Failed to get build properties, {}", e.getMessage());
            log.debug("Failed to get build properties", e);
        }
        return buildPropsList;
    }

    public Long getBuildId(BuildParams buildParams) {
        Object[] diffParams = getBuildPropsParam(buildParams);
        try (ResultSet rs = jdbcHelper.executeSelect(BuildQueries.BUILD_BY_NAME_NUMBER_DATE, diffParams)) {
            if (rs.next()) {
                Long res = rs.getLong(1);
                if (rs.next()) {
                    log.error("More then one build by number: {}  date: {}", diffParams[0], diffParams[1]);
                }
                return res;
            }
        } catch (SQLException e) {
            log.error("Failed to retrieve buildId, {}", e.getMessage());
            log.debug("Failed to retrieve buildId", e);
        }
        return null;
    }

    /**
     * get build props (env or system) param to be included in sql query
     *
     * @param buildParams - build params
     * @return list of build props param
     */
    private Object[] getBuildPropsParam(BuildParams buildParams) {
        return new Object[] {buildParams.getBuildName(), buildParams.getCurrBuildNum(), buildParams.getCurrBuildDate()};
    }

    /**
     * get Module Artifact diff with paging
     */
    public List<ModuleDependency> getModuleDependencyForDiffWithPaging(BuildParams buildParams) {
        List<ModuleDependency> dependencies = new ArrayList<>();
        Object[] diffParams = getBuildDependencyParams(buildParams);
        try (ResultSet rs = jdbcHelper.executeSelect(BuildQueries.MODULE_DEPENDENCY_DIFF_QUERY, diffParams)) {
            Map<String, String> tempDependencyMap = new HashMap<>();
            StringBuilder inClauseBuilder = new StringBuilder();
            inClauseBuilder.append("(");
            while (rs.next()) {
                String sha1 = rs.getString(3);
                if (tempDependencyMap.get(sha1) == null) {
                    tempDependencyMap.put(sha1, sha1);
                    ModuleDependency dependency = new ModuleDependency(null, null, rs.getString(1),
                            rs.getString(2), rs.getString(4), sha1);
                    dependency.setStatus(rs.getString(5));
                    dependencies.add(dependency);
                }
                inClauseBuilder.append("'").append(sha1).append("'").append(",");
            }
            updateDependenciesRepoPathData(dependencies, inClauseBuilder);
        } catch (SQLException e) {
            log.error("Failed to get module dependencies with paging, {}", e.getMessage());
            log.debug("Failed to get module dependencies with paging", e);
        }
        return dependencies;
    }

    private void updateDependenciesRepoPathData(List<ModuleDependency> dependencies, StringBuilder inClauseBuilder)
            throws SQLException {
        String inClause = inClauseBuilder.toString();
        inClause = inClause.substring(0, inClause.length() - 1);
        inClause = inClause + ")";
        // update dependencies repo path data
        if (!dependencies.isEmpty()) {
            Map<String, ModuleDependency> moduleDependencyMap = getModuleDependencyNodes(inClause);
            dependencies.forEach(dependency -> {
                ModuleDependency moduleDependency = moduleDependencyMap.get(dependency.getSha1());
                if (moduleDependency != null) {
                    dependency.setRepoKey(moduleDependency.getRepoKey());
                    String path = moduleDependency.getPath();
                    String name = moduleDependency.getName();
                    if (path != null) {
                        dependency.setPath(path.equals(".") ? name : path + "/" + name);
                    }
                }
            });
        }
    }

    /**
     * get build dependency query params for all build dependency diff or build module dependency diff
     *
     * @param buildParams - build diff param
     * @return - build dependency param for diff query
     */
    private Object[] getBuildDependencyParams(BuildParams buildParams) {
        return new Object[] {
                buildParams.getCurrBuildNum(), buildParams.getCurrBuildDate(), buildParams.getBuildRepo(),
                buildParams.getBuildModuleId(), /// ===
                buildParams.getComperedBuildNum(), buildParams.getComperedBuildDate(), buildParams.getBuildRepo(),
                buildParams.getBuildModuleId(), /// ===
                buildParams.getComperedBuildNum(), buildParams.getComperedBuildDate(), buildParams.getBuildRepo(),
                buildParams.getBuildModuleId(), /// ===
                buildParams.getCurrBuildNum(), buildParams.getCurrBuildDate(), buildParams.getBuildRepo(),
                buildParams.getBuildModuleId(), /// ===
                buildParams.getCurrBuildNum(), buildParams.getCurrBuildDate(), buildParams.getBuildRepo(),
                buildParams.getBuildModuleId(), /// ===
                buildParams.getComperedBuildNum(), buildParams.getComperedBuildDate(), buildParams.getBuildRepo(),
                buildParams.getBuildModuleId(), /// ===
                buildParams.getCurrBuildNum(), buildParams.getCurrBuildDate(), buildParams.getBuildRepo(),
                buildParams.getBuildModuleId(), /// ===
                buildParams.getComperedBuildNum(), buildParams.getComperedBuildDate(), buildParams.getBuildRepo(),
                buildParams.getBuildModuleId()}; /// ===
    }

    public long findBuildId(String name, String number, long startDate) throws SQLException {
        try (ResultSet rs = jdbcHelper.executeSelect(
                "SELECT build_id FROM builds WHERE build_name = ? AND build_number = ? AND build_date = ?", name,
                number, startDate)) {
            if (rs.next()) {
                return rs.getLong(1);
            }
        }
        return 0L;
    }

    public BuildEntity findBuild(String name, String number, long startDate) throws SQLException {
        long buildId = findBuildId(name, number, startDate);
        if (buildId > 0L) {
            return getBuild(buildId);
        }
        return null;
    }

    public BuildEntity getLatestBuild(String buildName, String buildRepo) throws SQLException {
        long latestBuildDate = 0L;
        try (ResultSet rs = jdbcHelper.executeSelect("SELECT max(build_date) FROM builds WHERE build_name = ? " +
                "AND repo = ?", buildName, buildRepo)) {
            if (rs.next()) {
                latestBuildDate = rs.getLong(1);
            }
        }

        BuildEntity buildEntity = null;
        if (latestBuildDate > 0L) {
            try (ResultSet rs = jdbcHelper.executeSelect("SELECT * FROM builds " +
                    "WHERE build_name = ? AND build_date = ? AND repo = ?", buildName, latestBuildDate, buildRepo)) {
                if (rs.next()) {
                    buildEntity = resultSetToBuild(rs);
                }
            }
        }
        if (buildEntity != null) {
            buildEntity.setProperties(findBuildProperties(buildEntity.getBuildId()));
            buildEntity.setPromotions(findBuildPromotions(buildEntity.getBuildId()));
        }
        return buildEntity;
    }

    public BuildRun findLatestBuild(String buildName, String buildNumber, String repoKey) throws SQLException {
        long latestBuildDate = 0L;
        try (ResultSet rs = jdbcHelper.executeSelect("SELECT max(build_date) FROM builds WHERE" +
                        " build_name = ? AND build_number = ? AND repo = ?",
                buildName, buildNumber, repoKey)) {
            if (rs.next()) {
                latestBuildDate = rs.getLong(1);
            }
        }
        if (latestBuildDate > 0L) {
            return new BuildRunImpl(buildName, buildNumber, BuildStoreServiceImpl.formatDateToString(latestBuildDate),
                    repoKey);
        }
        return null;
    }

    public List<Long> findBuildIds(String buildName, @Nonnull String buildRepo) throws SQLException {
        List<Long> buildIds = new ArrayList<>();
        try (ResultSet rs = jdbcHelper.executeSelect("SELECT build_id FROM builds WHERE" +
                " build_name = ? AND repo = ? ORDER BY build_date DESC", buildName, buildRepo)) {
            while (rs.next()) {
                buildIds.add(rs.getLong(1));
            }
        }
        return buildIds;
    }

    public List<Long> findBuildIds(String buildName, String buildNumber, @Nonnull String buildRepo)
            throws SQLException {
        List<Long> buildIds = new ArrayList<>();
        try (ResultSet rs = jdbcHelper.executeSelect("SELECT build_id FROM builds WHERE" +
                        " build_name = ? AND build_number = ? AND repo = ? ORDER BY build_date DESC",
                buildName, buildNumber, buildRepo)) {
            while (rs.next()) {
                buildIds.add(rs.getLong(1));
            }
        }
        return buildIds;
    }

    public List<BuildNameAndRepo> getAllBuildNames(@Nullable String buildRepo) throws SQLException {
        ResultSet rs = null;
        List<BuildNameAndRepo> buildNames = new ArrayList<>();
        try {
            if (StringUtils.isNotBlank(buildRepo)) {
                rs = jdbcHelper.executeSelect(
                        "SELECT build_name, repo, max(build_date) d FROM builds WHERE repo = ? GROUP BY build_name, " +
                                "repo ORDER BY d", buildRepo);
            } else {
                rs = jdbcHelper.executeSelect(
                        "SELECT build_name, repo, max(build_date) d FROM builds GROUP BY build_name, repo ORDER BY d");
            }
            while (rs.next()) {
                buildNames.add(new BuildNameAndRepo(rs.getString(BUILD_NAME_FIELD), rs.getString("repo")));
            }
        } finally {
            DbUtils.close(rs);
        }
        return buildNames;
    }

    @NotNull
    public int getBuildsCountFromSpecificRepo(String repo) throws SQLException {
        return jdbcHelper.executeSelectCount("SELECT COUNT(*) FROM builds WHERE repo = ?", repo);
    }

    @NotNull
    public int getTotalBuildsCount() throws SQLException {
        return jdbcHelper.executeSelectCount("SELECT COUNT(*) FROM builds");
    }

    public String getBuildStoringRepo(@NotNull String name, @NotNull String number, long date)
            throws SQLException {
        String repo = null;
        try (ResultSet rs = jdbcHelper.executeSelect(
                "SELECT repo FROM builds WHERE build_name = ? AND build_number = ? AND build_date = ?",
                name, number, date)) {
            if (rs.next()) {
                repo = rs.getString(1);
            }
        }
        return repo;
    }

    /**
     * get build modules with paging
     *
     * @param buildNumber - build name
     * @param date        - date
     */
    public List<PublishedModule> getBuildModule(String buildNumber, String date, String buildRepo)
            throws SQLException {
        String buildQuery = "SELECT build_modules.module_name_id,\n" +
                "(select count(*) from build_artifacts where build_artifacts.module_id =  build_modules.module_id ) " +
                "as num_of_art ,\n" +
                "(select count(*) from build_dependencies where build_dependencies.module_id =  build_modules" +
                ".module_id ) as num_of_dep FROM build_modules\n" +
                "left join builds on builds.build_id=build_modules.build_id \n" +
                "where  builds.build_number=? and builds.build_date=? AND builds.repo = ?";
        List<PublishedModule> modules = new ArrayList<>();
        try (ResultSet rs = jdbcHelper.executeSelect(buildQuery, buildNumber, Long.parseLong(date), buildRepo)) {
            while (rs.next()) {
                PublishedModule module = new PublishedModule();
                module.setId(rs.getString(1));
                module.setNumOfArtifact(rs.getString(2));
                module.setNumOfDependencies(rs.getString(3));
                modules.add(module);
            }
        }
        return modules;
    }

    /**
     * get build modules artifact with paging
     *
     * @param buildNumber - build name
     * @param date        - build date
     * @param onlyDefaultBuildRepoExists - true if only:
     * {@link org.artifactory.addon.build.BuildAddon#BUILD_INFO_REPO_NAME} exists, false otherwise
     * @throws SQLException - in case of DB errors
     */
    public List<ModuleArtifact> getModuleArtifact(String buildName, String buildNumber, String date,
            String moduleId, String buildRepo, boolean onlyDefaultBuildRepoExists) throws SQLException {
        long buildDate = Long.parseLong(date);
        List<ModuleArtifact> artifacts = new ArrayList<>();
        try (ResultSet rs = getPaginatedArtifact(buildNumber, buildDate, moduleId, buildRepo)) {
            // get artifact info
            while (rs.next()) {
                artifacts.add(new ModuleArtifact(null, null, rs.getString(1), rs.getString(2), rs.getString(3)));
            }
            if (!artifacts.isEmpty()) {
                // query for artifact nodes
                final Map<String, ModuleArtifact> artifactNodes =
                        getArtifactNodes(buildName, buildNumber, buildDate, onlyDefaultBuildRepoExists);
                for (ModuleArtifact artifact : artifacts) {
                    ModuleArtifact moduleArtifact = artifactNodes.get(artifact.getSha1());
                    if (moduleArtifact != null) {
                        artifact.setRepoKey(moduleArtifact.getRepoKey());
                        artifact.setPath(moduleArtifact.getPath());
                    }
                }
            }
        }
        return artifacts;
    }

    /**
     * get module artifact info
     */
    private ResultSet getPaginatedArtifact(String buildNumber, Long date, String moduleId, String buildRepo)
            throws SQLException {
        ResultSet rs;
        String buildQuery =
                "SELECT distinct build_artifacts.artifact_name as name,build_artifacts.artifact_type as type," +
                        "build_artifacts.sha1 FROM build_artifacts\n" +
                        "left join build_modules on build_modules.module_id=build_artifacts.module_id\n" +
                        "left join builds on  build_modules.build_id = builds.build_id\n" +
                        "where builds.build_number = ? and builds.build_date = ? and builds.repo = ? and " +
                        "build_modules.module_name_id = ?";

        rs = jdbcHelper.executeSelect(buildQuery, buildNumber, date, buildRepo, moduleId);

        return rs;
    }

    /**
     * get Artifact nodes data by build name and number
     * Note - who ever use this method must be responsible for close the result set
     *
     * @return map of data , key = sha1 , value = module artifact
     * @throws SQLException - in case of DB errors
     */
    private Map<String, ModuleArtifact> getArtifactNodes(String buildName, String buildNumber, long timestamp,
            boolean onlyDefaultBuildRepoExists)
            throws SQLException {
        if (onlyDefaultBuildRepoExists) {
            return getArtifactNodesByNameAndNumber(buildName, buildNumber);
        }
        Map<String, ModuleArtifact> artifactMap =
                getArtifactNodesByBuildTimestamp(buildName, buildNumber, timestamp);

        if (artifactMap.isEmpty()) {
            return getArtifactNodesByNameAndNumber(buildName, buildNumber);
        }
        return artifactMap;
    }

    private Map<String, ModuleArtifact> getArtifactNodesByBuildTimestamp(String buildName, String buildNumber,
            long timestamp) throws SQLException {
        Map<String, ModuleArtifact> artifactMap = new HashMap<>();
        String query = BuildQueries.GET_ARTIFACT_NODES_BY_BUILD_TIMESTAMP;
        try (ResultSet rsArtifact = jdbcHelper
                .executeSelect(query, buildName, buildNumber, TrashService.TRASH_KEY, String.valueOf(timestamp))) {

            while (rsArtifact.next()) {
                String sha1 = rsArtifact.getString(6);
                if (artifactMap.get(sha1) == null) {
                    artifactMap.put(sha1, new ModuleArtifact(rsArtifact.getString(1), rsArtifact.getString(2),
                            rsArtifact.getString(3),
                            null, null));
                }
            }
        }
        return artifactMap;
    }

    private Map<String, ModuleArtifact> getArtifactNodesByNameAndNumber(String buildName, String buildNumber)
            throws SQLException {
        Map<String, ModuleArtifact> artifactMap = new HashMap<>();
        String query = BuildQueries.GET_ARTIFACT_NODES;
        try (ResultSet rsArtifact = jdbcHelper.executeSelect(query, buildName, buildNumber, TrashService.TRASH_KEY)) {
            while (rsArtifact.next()) {
                String sha1 = rsArtifact.getString(6);
                if (artifactMap.get(sha1) == null) {
                    artifactMap.put(sha1, new ModuleArtifact(rsArtifact.getString(1), rsArtifact.getString(2),
                            rsArtifact.getString(3),
                            null, null));
                }
            }
        }
        return artifactMap;
    }

    /**
     * get build modules dependencies with paging
     */
    public List<ModuleDependency> getModuleDependency(String buildNumber, String date, String buildRepo,
            String moduleId)
            throws SQLException {
        String buildQuery = "SELECT distinct build_dependencies.dependency_name_id as id," +
                "build_dependencies.dependency_type as type,build_dependencies.dependency_scopes as scope," +
                "build_dependencies.sha1 FROM build_dependencies\n" +
                "left join build_modules on build_modules.module_id=build_dependencies.module_id\n" +
                "left join builds on  build_modules.build_id = builds.build_id\n" +
                "where builds.build_number = ? and builds.build_date = ? and builds.repo = ? and build_modules" +
                ".module_name_id = ?";

        List<ModuleDependency> dependencies = new ArrayList<>();
        try (ResultSet rs = jdbcHelper.executeSelect(buildQuery, buildNumber, Long.parseLong(date),
                buildRepo, moduleId)) {
            StringBuilder inClauseBuilder = new StringBuilder();
            inClauseBuilder.append("(");
            while (rs.next()) {
                String sha1 = rs.getString(4);
                dependencies.add(new ModuleDependency(null, null, rs.getString(1), rs.getString(2), rs.getString(3),
                        sha1));
                inClauseBuilder.append("'").append(sha1).append("'").append(",");
            }
            updateDependenciesRepoPathData(dependencies, inClauseBuilder);
        }
        return dependencies;
    }

    private Map<String, ModuleDependency> getModuleDependencyNodes(String inClause) throws SQLException {
        Map<String, ModuleDependency> moduleDependencyMap = new HashMap<>();
        try (ResultSet rsDep = jdbcHelper.executeSelect(
                "SELECT distinct nodes.repo,nodes.node_path,nodes.node_name,nodes.sha1_actual FROM nodes\n" +
                        "                where nodes.repo != '" + TrashService.TRASH_KEY +
                        "' and nodes.sha1_actual in " + inClause)) {
            while (rsDep.next()) {
                String sha1 = rsDep.getString(4);
                if (moduleDependencyMap.get(sha1) == null) {
                    moduleDependencyMap.put(sha1,
                            new ModuleDependency(rsDep.getString(1), rsDep.getString(2), rsDep.getString(3), null, null,
                                    null));
                }
            }
        }
        return moduleDependencyMap;
    }

    public List<BuildIdEntity> getLatestBuildIds(ContinueBuildFilter continueBuildFilter) {
        ResultSet rs = null;
        List<BuildIdEntity> buildNames = new ArrayList<>();
        if (continueBuildFilter.getLimit() <= 0) {
            return buildNames;
        }
        try {
            QueryWriter queryWriter = new QueryWriter(dbService.getDatabaseType());
            queryWriter.select("d.build_id, d.build_name, d.build_number, d.build_date, d.repo")
                    .from("builds d ")
                    .innerJoin(
                            "(select c.build_name, max(c.build_date) build_time from builds c group by c.build_name, " +
                                    "c.repo ) k",
                            "k.build_name=d.build_name AND k.build_time=d.build_date")
                    .limit(continueBuildFilter.getLimit());
            String orderByStr = BuildsDaoUtils.createOrderByStr(continueBuildFilter);
            queryWriter.orderBy(orderByStr);
            List<Object> sqlParametersList = new ArrayList<>();
            String whereClause =
                    BuildsDaoUtils.createWhereClauseFromContinueBuildFilter(continueBuildFilter, sqlParametersList);
            if (StringUtils.isNotBlank(whereClause)) {
                queryWriter.where(whereClause);
            }
            rs = jdbcHelper.executeSelect(queryWriter.build(), sqlParametersList.toArray());
            buildNames = BuildsDaoUtils.populateBuildIdEntryList(rs);
        } catch (Exception e) {
            log.error("Failed to get latest builds ids, {}", e.getMessage());
            log.debug("Failed to get latest builds ids", e);
        } finally {
            DbUtils.close(rs);
        }
        return buildNames;
    }

    /**
     * Default sorting is by date desc
     */
    @Nonnull
    public List<BuildIdEntity> findLatestBuildIdsByNameCaseInsensitive(@Nullable String buildRepo, String buildName,
            String fromDate, String toDate,
            String orderBy, String direction) throws SQLException {
        ResultSet rs = null;
        List<BuildIdEntity> builds = new ArrayList<>();
        List<Object> whereParams = new LinkedList<>();
        try {
            String query = "SELECT a.build_id, a.build_name, a.build_number, a.build_date, a.repo\n" +
                    "   FROM builds a\n" +
                    "   INNER JOIN\n" +
                    "     (SELECT c.build_name, c.repo, max(build_date) build_time\n" +
                    "     FROM builds c\n" +
                    "     INNER JOIN\n" +
                    "           (SELECT build_name, repo\n" +
                    "            FROM builds\n" +
                    buildILikeWhereClause(whereParams, buildRepo, buildName, fromDate, toDate) +
                    "            GROUP BY build_name, repo) d\n" +
                    "     ON c.build_name = d.build_name AND c.repo = d.repo\n" +
                    "     GROUP BY c.build_name, c.repo) b\n" +
                    "   ON a.build_name = b.build_name AND a.build_date = build_time AND a.repo = b.repo\n" +
                    buildOrderByClause(orderBy, direction);
            rs = jdbcHelper.executeSelect(query, whereParams);
            while (rs.next()) {
                builds.add(new BuildIdEntity(rs.getLong(1), rs.getString(2), rs.getString(3), rs.getLong(4),
                        rs.getString(5)));
            }
        } finally {
            DbUtils.close(rs);
        }
        return builds;
    }

    private String buildILikeWhereClause(List<Object> whereParams, @Nullable String buildRepo, String buildName,
            String fromDate, String toDate) {
        StringBuilder where = new StringBuilder();
        if (StringUtils.isNotBlank(buildName)) {
            String name = buildName.replace("?", "_").replace("*", "%");
            where.append("lower(build_name) LIKE ?");
            whereParams.add(name.toLowerCase() + "%");
        }
        if (StringUtils.isNotBlank(buildRepo)) {
            if (where.length() > 0) {
                where.append(AND);
            }
            where.append("repo = ?");
            whereParams.add(buildRepo);
        }
        if (StringUtils.isNotBlank(fromDate)) {
            if (where.length() > 0) {
                where.append(AND);
            }
            where.append("build_date >= ?");
            whereParams.add(Long.valueOf(fromDate));
        }
        if (StringUtils.isNotBlank(toDate)) {
            if (where.length() > 0) {
                where.append(AND);
            }
            where.append("build_date <= ?");
            whereParams.add(Long.valueOf(toDate));
        }
        return where.length() > 0 ? " WHERE " + where.toString() : "";
    }

    private String buildOrderByClause(String orderBy, String direction) throws SQLException {
        String sortingDirection = "asc".equalsIgnoreCase(direction) ? "ASC" : "DESC";
        StringBuilder orderClause = new StringBuilder();
        if (StringUtils.isNotBlank(orderBy)) {
            switch (orderBy) {
                case "name":
                    orderClause.append(BUILD_NAME_FIELD);
                    break;
                case "number":
                    orderClause.append("build_number");
                    break;
                case "date":
                    orderClause.append(BUILD_DATE_FIELD);
                    break;
                default:
                    throw new SQLException("Cannot order by: " + orderBy + ". Unknown column.");
            }
        } else {
            orderClause.append(BUILD_DATE_FIELD);
        }
        return " ORDER BY a." + orderClause.append(" ").append(sortingDirection).toString();
    }

    public Collection<BuildEntity> findBuildsForArtifactChecksum(ChecksumType type, String checksum, String buildRepo)
            throws SQLException {
        return findBuildsByArtifactChecksum(type, checksum, "build_artifacts", buildRepo);
    }

    public Collection<BuildEntity> findBuildsForDependencyChecksum(ChecksumType type, String checksum, String buildRepo)
            throws SQLException {
        return findBuildsByArtifactChecksum(type, checksum, "build_dependencies", buildRepo);
    }

    private Collection<BuildEntity> findBuildsByArtifactChecksum(ChecksumType type, String checksum,
            String artifactsTable, String buildRepo) throws SQLException {

        Collection<BuildEntity> results = Lists.newArrayList();
        if (ChecksumType.sha256.equals(type)) {
            // Kicked out build artifacts sha2 from db because of performance
            log.debug("Build artifact sha256 search not supported.");
            return results;
        }

        List<Object> params = new ArrayList<>();
        String query = "SELECT DISTINCT b.* FROM builds b, " + artifactsTable +
                " artifactsTable, build_modules bm" +
                " WHERE b.build_id = bm.build_id" +
                " AND bm.module_id = artifactsTable.module_id" +
                " AND artifactsTable." + type.name() + " = ?" +
                " AND artifactsTable.module_id = bm.module_id" +
                " AND b.repo = ?";

        params.add(checksum);
        params.add(buildRepo);

        try (ResultSet rs = jdbcHelper.executeSelect(query, params)) {
            while (rs.next()) {
                results.add(resultSetToBuild(rs));
            }
        }
        for (BuildEntity buildEntity : results) {
            buildEntity.setProperties(findBuildProperties(buildEntity.getBuildId()));
            buildEntity.setPromotions(findBuildPromotions(buildEntity.getBuildId()));
        }
        return results;
    }

    private Set<BuildProperty> findBuildProperties(long buildId) throws SQLException {
        Set<BuildProperty> buildProperties = new HashSet<>();
        try (ResultSet rs = jdbcHelper.executeSelect("SELECT * FROM build_props WHERE build_id = ?", buildId)) {
            while (rs.next()) {
                buildProperties.add(resultSetToBuildProperty(rs));
            }
        }
        return buildProperties;
    }

    private SortedSet<BuildPromotionStatus> findBuildPromotions(long buildId) throws SQLException {
        SortedSet<BuildPromotionStatus> buildPromotions = new TreeSet<>();
        try (ResultSet rs = jdbcHelper.executeSelect("SELECT * FROM build_promotions WHERE build_id = ?", buildId)) {
            while (rs.next()) {
                buildPromotions.add(resultSetToBuildPromotion(rs));
            }
        }
        return buildPromotions;
    }

    private static BuildProperty resultSetToBuildProperty(ResultSet rs) throws SQLException {
        return new BuildProperty(rs.getLong(1), rs.getLong(2),
                rs.getString(3), rs.getString(4));
    }

    private static BuildProps resultSetToBuildProps(ResultSet rs) throws SQLException {
        return new BuildProps(rs.getString(1), rs.getString(2), null);
    }

    private BuildPromotionStatus resultSetToBuildPromotion(ResultSet rs) throws SQLException {
        return new BuildPromotionStatus(rs.getLong(1), rs.getLong(2),
                rs.getString(3), rs.getString(4), rs.getString(5), rs.getString(6),
                rs.getString(7));
    }

    private BuildEntity resultSetToBuild(ResultSet rs) throws SQLException {
        return new BuildEntity(rs.getLong(1),
                rs.getString(2), rs.getString(3), rs.getLong(4),
                rs.getString(5), rs.getLong(6), rs.getString(7),
                zeroIfNull(rs.getLong(8)), rs.getString(9), rs.getString(10)
        );
    }

    private GeneralBuild resultSetToGeneralBuild(ResultSet rs, Long id) throws SQLException {
        return new GeneralBuild(id,
                rs.getString(BUILD_NAME_FIELD), rs.getString("build_number"), rs.getLong(BUILD_DATE_FIELD),
                rs.getString("ci_url"), rs.getLong("created"), rs.getString("created_by"),
                zeroIfNull(rs.getLong("modified")), rs.getString("modified_by"), rs.getString("repo")
        );
    }

    /**
     * Used to determine how many entries in the build_jsons table (for the migration job)
     */
    public int getNumberOfBuildJsons() throws SQLException {
        int num = 0;
        if (tableExists(jdbcHelper, dbProperties.getDbType(), BUILD_JSONS_TABLE)) {
            num = jdbcHelper.executeSelectCount("SELECT COUNT(*) FROM " + BUILD_JSONS_TABLE);
        }
        return num;
    }

    /**
     * Gets the next {@param limit} build ids ordered (for the migration job)
     */
    public List<Long> getBuildIdsStartingAt(long lastBuildId, int limit) throws SQLException {
        List<Long> buildIds = new ArrayList<>();
        String query = new ArtifactoryQueryWriter().select("build_id ")
                .from(BUILD_JSONS_TABLE + " ")
                .where("build_id > ? ")
                .orderBy("build_id ASC ")
                .limit((long) limit)
                .build();
        try (ResultSet rs = jdbcHelper.executeSelect(query, lastBuildId)) {
            while (rs.next()) {
                buildIds.add(rs.getLong("build_id"));
            }
        }
        return buildIds;
    }
}
