package org.artifactory.storage.db.conversion.version.v277;

import com.google.common.annotations.VisibleForTesting;
import org.artifactory.common.ConstantValues;
import org.artifactory.storage.db.version.converter.DBSqlConverter;
import org.artifactory.storage.db.version.converter.DbSqlConverterUtil;
import org.jfrog.storage.DbType;
import org.jfrog.storage.JdbcHelper;
import org.jfrog.storage.util.DbUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static org.jfrog.storage.util.DbUtils.doWithConnection;

/**
 * A converter with 2 phases in 1 TX:
 * 1. static script that enhanced node_events and node_events_error and creates node_events_tmp
 * 2. dynamic script SQL that creates event_id sequence with dynamic start value
 * @author shaharz
 */

public class V227NodeEventsAndTmpAndErrorsDBSqlConverter extends DBSqlConverter {

    private static final Logger log = LoggerFactory.getLogger(V227NodeEventsAndTmpAndErrorsDBSqlConverter.class);

    private static final String CREATE_SEQ_MSSQL_DERBY = "CREATE SEQUENCE node_events_event_id_seq " +
            "AS BIGINT START WITH %1$s";
    private static final String CREATE_SEQ_PG = "CREATE SEQUENCE node_events_event_id_seq START WITH %1$s";

    // In oracle the default precision is number of 28 digits, limiting it to 19.
    private static final String CREATE_SEQ_ORA = "CREATE SEQUENCE node_events_event_id_seq " +
            "START WITH %1$s MAXVALUE 9999999999999999999";

    // In MsSql we can't use the sequence explicitly in as Insert Select,
    // hence setting it as default value of event_id column
    private static final String ALTER_NODE_EVENTS_EVENT_ID_DEF_AS_SEQ_MSSQL = "ALTER TABLE node_events ADD " +
            "DEFAULT NEXT VALUE FOR node_events_event_id_seq FOR event_id";

    private static final String CURRENT_ID_QUERY = "SELECT current_id FROM unique_ids";

    private final String fromVersion;

    public V227NodeEventsAndTmpAndErrorsDBSqlConverter(String fromVersion) {
        super(fromVersion);
        this.fromVersion = fromVersion;
    }

    @Override
    public void convert(JdbcHelper jdbcHelper, DbType dbType){

        try {
            //Ensure that the script was never applied before
            if (DbUtils.tableExists(jdbcHelper, dbType, "node_events_tmp")){
                return;
            }

            //Invoke the static script and dynamic script in one transaction.
            doWithConnection(jdbcHelper, conn -> {
                conn.setAutoCommit(false);
                try {
                    //First invoke the static script
                    DbSqlConverterUtil.convert(conn, dbType, fromVersion);

                    //Now invoke the dynamic script
                    // In MYSQL/MARIADB AUTO INCREMENT was defined in 'mysql_v227_node_events_and_errors.sql'
                    // AUTO INCREMENT we don't have to create it with start value
                    if (!dbType.equals(DbType.MYSQL) && !dbType.equals(DbType.MARIADB)) {
                        try (Statement statement = conn.createStatement()) {
                            createSequence(dbType, start(statement), statement, conn);
                        }
                    }
                    conn.commit();
                } catch (Exception e) {
                    conn.rollback();
                    throw e;
                } finally {
                    conn.setAutoCommit(true);
                }
            });
        } catch (SQLException e) {
            //conversion error is runtime and logged in the util, no need to catch it here.
            String msg = "Could not convert DB: CREATE SEQUENCE node_events_event_id_seq";
            log.error(msg + " due to " + e.getMessage(), e);
            throw new RuntimeException(msg, e);
        }
    }


    @VisibleForTesting
    void createSequence(DbType dbType, long start, Statement statement, Connection conn) throws SQLException {
        List<String> sqls;
        switch (dbType) {
            case DERBY:
                sqls = Collections.singletonList(String.format(CREATE_SEQ_MSSQL_DERBY, start));
                break;
            case MSSQL:
                sqls = Arrays.asList(String.format(CREATE_SEQ_MSSQL_DERBY, start),
                        ALTER_NODE_EVENTS_EVENT_ID_DEF_AS_SEQ_MSSQL);
                break;
            case ORACLE:
                sqls = Collections.singletonList(String.format(CREATE_SEQ_ORA, start));
                break;
            case POSTGRESQL:
                sqls = Collections.singletonList(String.format(CREATE_SEQ_PG, start));
                break;
            default:
                throw new UnsupportedOperationException();
        }

        try {
            for (String sql : sqls) {
                statement.executeUpdate(sql);
            }
            conn.commit();
        } catch (SQLException e) {
            conn.rollback();
            throw e;
        }
    }

    private long start(Statement statement) throws SQLException {
        long start = 1;
        try (ResultSet resultSet = statement.executeQuery(CURRENT_ID_QUERY)) {
            if (resultSet.next()) {
                start = resultSet.getLong(1);
            }
        }
        return start + ConstantValues.nodeEventsEventIdMigrationMargin.getLong();
    }
}
