package org.artifactory.webapp.servlet;

import com.google.common.collect.ImmutableList;
import org.artifactory.opentracing.ArtifactoryTracer;
import org.artifactory.opentracing.ServerTracingSpan;
import org.jfrog.common.opentracing.OpenTracerBuilder;
import org.slf4j.MDC;
import org.testng.annotations.Test;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import static org.artifactory.opentracing.ArtifactoryTracer.UBER_TRACE_ID;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.argThat;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;

/**
 * @author Uriah Levy
 */
@Test
public class ArtifactoryTracingFilterTest {

    public void testFilterWithIncomingTraceHeader() throws IOException, ServletException {
        ArtifactoryTracingFilter artifactoryTracingFilter = new ArtifactoryTracingFilter();
        ArtifactoryTracer tracerSpy = spy(new ArtifactoryTracer(
                OpenTracerBuilder.getQuietJaegerTracer("Artifactory Tracer")));
        artifactoryTracingFilter.setArtifactoryTracer(tracerSpy);

        HttpServletRequest request = mock(HttpServletRequest.class);
        when(request.getHeaderNames()).thenReturn(Collections.enumeration(ImmutableList.of(UBER_TRACE_ID)));
        String fullTraceId = "63fdd1725a3c5673:276415bdb255cfe1:62c7e48c60a7afba:0";
        when(request.getHeader(UBER_TRACE_ID)).thenReturn(fullTraceId);
        String requestPath = "artifactory/api/system/ping";
        when(request.getPathInfo()).thenReturn(requestPath);

        FilterChain filterChain = mock(FilterChain.class);
        List<String> shortTraceIdHolder = new ArrayList<>();
        doAnswer((x) -> shortTraceIdHolder.add(MDC.get(UBER_TRACE_ID))).when(filterChain).doFilter(any(), any());

        artifactoryTracingFilter.doFilter(request, mock(ServletResponse.class), filterChain);
        verify(tracerSpy)
                .startSpanOnCurrentThread(argThat(requestPath::equals),
                        argThat(arg -> arg.values().stream().anyMatch(fullTraceId::equals)));
        // Original trace ID preserved
        assertEquals("63fdd1725a3c5673", shortTraceIdHolder.get(0));
        // Span is finished
        assertNull(tracerSpy.getCurrentThreadActiveSpan());
        assertNull(MDC.get(UBER_TRACE_ID));
    }

    public void testFilterWithoutIncomingTraceHeader() throws IOException, ServletException {
        ArtifactoryTracingFilter artifactoryTracingFilter = new ArtifactoryTracingFilter();
        ArtifactoryTracer tracerSpy = spy(new ArtifactoryTracer(
                OpenTracerBuilder.getQuietJaegerTracer("Artifactory Tracer")));
        artifactoryTracingFilter.setArtifactoryTracer(tracerSpy);

        HttpServletRequest request = mock(HttpServletRequest.class);
        when(request.getHeaderNames()).thenReturn(Collections.enumeration(ImmutableList.of("Host")));
        when(request.getHeader("Host")).thenReturn("foo.com");
        String requestPath = "artifactory/api/system/ping";
        when(request.getPathInfo()).thenReturn(requestPath);

        FilterChain filterChain = mock(FilterChain.class);
        List<String> shortTraceIdHolder = new ArrayList<>();
        List<ServerTracingSpan> tracingSpanHolder = new ArrayList<>();
        doAnswer((x) -> {
                    shortTraceIdHolder.add(MDC.get(UBER_TRACE_ID));
                    return tracingSpanHolder.add(tracerSpy.getCurrentThreadActiveSpan());
                }
        ).when(filterChain).doFilter(any(), any());

        artifactoryTracingFilter.doFilter(request, mock(ServletResponse.class), filterChain);
        verify(tracerSpy)
                .startSpanOnCurrentThread(argThat(requestPath::equals),
                        argThat(arg -> arg.values().stream().anyMatch(x -> true)));
        // Span is null
        assertNull(tracerSpy.getCurrentThreadActiveSpan());
        // A new span was created
        assertNotNull(tracingSpanHolder.get(0));
        assertEquals(tracingSpanHolder.get(0).getSpan().context().toTraceId(), shortTraceIdHolder.get(0));
    }
}