package org.artifactory.ui.rest.service.admin.configuration.repositories.util.validator;

import com.google.common.collect.Maps;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.rest.common.exception.RepoConfigException;
import org.artifactory.test.ArtifactoryHomeStub;
import org.artifactory.ui.rest.model.admin.configuration.repository.GeneralRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteAdvancedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteBasicRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteCacheRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.DockerTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.GenericTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.GoTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.TypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.virtual.VirtualSelectedRepository;
import org.junit.Before;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.validation.constraints.NotNull;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_DOCKER_REMOTE_RETRIEVAL_CACHE_PERIOD;
import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_GO_REMOTE_METADATA_MISSED_RETRIEVAL_CACHE_PERIOD;
import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_MISSED_RETRIEVAL_PERIOD;
import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_RETRIEVAL_CACHE_PERIOD;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

/**
 * @author AndreiK.
 */
@Test
public class RepoConfigValidatorTest {

    private RepoConfigValidator validator;

    @Before
    public void setup() {
        MockitoAnnotations.initMocks(this);
        ArtifactoryHome.bind(new ArtifactoryHomeStub());
        validator = new RepoConfigValidator();
    }

    @Test(dataProvider = "provideDockerCachePeriodAndRepoType", description =
            "Actually sets the value when changed in the" +
                    " UI note the cast to DockerTypeSpecificConfigModel - this replicates the UI originated model")
    public void testRemoteRepoValidationOfRetrievalCachePeriodSecsOfCacheModel(Long cachePeriod, boolean isDockerModel,
            Long expected) {
        TypeSpecificConfigModel typeSpecificConfigModel;
        if (isDockerModel) {
            typeSpecificConfigModel = new DockerTypeSpecificConfigModel();
            ((DockerTypeSpecificConfigModel) typeSpecificConfigModel).setRetrievalCachePeriodSecs(cachePeriod);
        } else {
            typeSpecificConfigModel = new GenericTypeSpecificConfigModel();
        }
        RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel =
                getAdvancedRepoConfigModelAndSetRetrievalCachePeriod(cachePeriod);
        RemoteRepositoryConfigModel model = getRemoteRepositoryConfigModel(remoteAdvancedRepositoryConfigModel,
                typeSpecificConfigModel);
        RemoteCacheRepositoryConfigModel cache = model.getAdvanced().getCache();

        validator.setRetrievalCachePeriodSecsOfCacheModel(model, cache);

        assertEquals((long) cache.getRetrievalCachePeriodSecs(), (long) expected);
    }

    @Test(dataProvider = "provideGoCachePeriodAndRepoType")
    public void testRemoteRepoValidationOfGoRetrievalCachePeriodSecsOfCacheModel(Long cachePeriod, boolean isGoModel,
            Long expected) {
        TypeSpecificConfigModel typeSpecificConfigModel;
        if (isGoModel) {
            typeSpecificConfigModel = new GoTypeSpecificConfigModel();
            ((GoTypeSpecificConfigModel) typeSpecificConfigModel)
                    .setMissedRetrievalCachePeriodSecs(cachePeriod);
        } else {
            typeSpecificConfigModel = new GenericTypeSpecificConfigModel();
        }
        RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel =
                getAdvancedRepoConfigModelAndSetMissedRetrievalCachePeriod(cachePeriod);
        RemoteRepositoryConfigModel model = getRemoteRepositoryConfigModel(remoteAdvancedRepositoryConfigModel,
                typeSpecificConfigModel);
        RemoteCacheRepositoryConfigModel cache = model.getAdvanced().getCache();

        validator.setRetrievalCachePeriodSecsOfCacheModel(model, cache);

        assertEquals((long) cache.getMissedRetrievalCachePeriodSecs(), (long) expected);
    }

    @Test
    public void testMapRepoKeyToDescriptor() {
        String repo = "repo";
        VirtualSelectedRepository repository = new VirtualSelectedRepository();
        repository.setRepoName(repo);
        RepoConfigValidator configValidator = new RepoConfigValidator();
        RepoDescriptor repoDescriptor =
                configValidator.mapRepoKeyToDescriptor(repository, createMutableDescriptorMock(repo));
        assertEquals(repoDescriptor.getKey(), repo);
    }

    @Test
    public void testValidateSelectedReposInVirtualExist() throws RepoConfigException {
        String repo = "repo";
        List<VirtualSelectedRepository> repoKeys = new ArrayList<>();
        VirtualSelectedRepository repository = new VirtualSelectedRepository();
        repository.setRepoName(repo);
        repoKeys.add(repository);
        RepoConfigValidator configValidator = new RepoConfigValidator();
        MutableCentralConfigDescriptor mutableDescriptorMock = createMutableDescriptorMock(repo);
        configValidator.validateSelectedReposInVirtualExist(repoKeys, mutableDescriptorMock);
        verify(mutableDescriptorMock, times(1)).getFederatedRepositoriesMap();
    }

    private MutableCentralConfigDescriptor createMutableDescriptorMock(String repoName) {
        MutableCentralConfigDescriptor descriptor = mock(MutableCentralConfigDescriptor.class);
        Map<String, FederatedRepoDescriptor> reposMap = Maps.newLinkedHashMap();
        FederatedRepoDescriptor federatedRepo = new FederatedRepoDescriptor();
        federatedRepo.setKey(repoName);
        reposMap.put(repoName, federatedRepo);
        when(descriptor.getFederatedRepositoriesMap()).thenReturn(reposMap);
        return descriptor;
    }

    @DataProvider
    private Object[][] provideDockerCachePeriodAndRepoType() {
        return new Object[][] {
                {null, true, DEFAULT_DOCKER_REMOTE_RETRIEVAL_CACHE_PERIOD},
                {null, false, DEFAULT_RETRIEVAL_CACHE_PERIOD},
                {36000L, true, 36000L},
                {36000L, false, 36000L},
        };
    }

    @DataProvider
    private Object[][] provideGoCachePeriodAndRepoType() {
        return new Object[][] {
                {null, true, DEFAULT_GO_REMOTE_METADATA_MISSED_RETRIEVAL_CACHE_PERIOD},
                {null, false, DEFAULT_MISSED_RETRIEVAL_PERIOD},
                {36000L, true, 36000L},
                {36000L, false, 36000L},
        };
    }

    @NotNull
    private RemoteAdvancedRepositoryConfigModel getAdvancedRepoConfigModelAndSetRetrievalCachePeriod(Long cachePeriod) {
        RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel =
                new RemoteAdvancedRepositoryConfigModel();
        RemoteCacheRepositoryConfigModel remoteCacheRepositoryConfigModel = new RemoteCacheRepositoryConfigModel();
        remoteAdvancedRepositoryConfigModel.setCache(remoteCacheRepositoryConfigModel);
        remoteCacheRepositoryConfigModel.setRetrievalCachePeriodSecs(cachePeriod);
        return remoteAdvancedRepositoryConfigModel;
    }

    @NotNull
    private RemoteAdvancedRepositoryConfigModel getAdvancedRepoConfigModelAndSetMissedRetrievalCachePeriod(
            Long cachePeriod) {
        RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel =
                new RemoteAdvancedRepositoryConfigModel();
        RemoteCacheRepositoryConfigModel remoteCacheRepositoryConfigModel = new RemoteCacheRepositoryConfigModel();
        remoteAdvancedRepositoryConfigModel.setCache(remoteCacheRepositoryConfigModel);
        remoteCacheRepositoryConfigModel.setMissedRetrievalCachePeriodSecs(cachePeriod);
        return remoteAdvancedRepositoryConfigModel;
    }

    @NotNull
    private RemoteRepositoryConfigModel getRemoteRepositoryConfigModel(
            RemoteAdvancedRepositoryConfigModel remoteAdvancedRepositoryConfigModel,
            TypeSpecificConfigModel typeSpecificConfigModel) {
        RemoteRepositoryConfigModel remoteRepositoryConfigModel = new RemoteRepositoryConfigModel();
        remoteRepositoryConfigModel.setBasic(new RemoteBasicRepositoryConfigModel());
        remoteRepositoryConfigModel.setGeneral(new GeneralRepositoryConfigModel());
        remoteRepositoryConfigModel.setAdvanced(remoteAdvancedRepositoryConfigModel);
        remoteRepositoryConfigModel.setTypeSpecific(typeSpecificConfigModel);
        return remoteRepositoryConfigModel;
    }
}